@@ -39,7 +39,8 @@ extern char *strsep();
 #endif /* not STDC_HEADERS */
 
 #define NLS_SET	db_deleteSet
-#include "nls/nls.h"
+#include <libintl.h>
+#define _(String) gettext (String)
 
 #include "manconfig.h"
 #include "lib/error.h"
@@ -100,7 +101,7 @@ int dbdelete(char *name, struct mandata *info)
 		multi_key = make_multi_key(name, *e);
 		if (!MYDBM_EXISTS(dbf, multi_key)) {
 			error (0, 0,
-			       CATGETS(db_delete_NO_MULTI, "multi key %s does not exist"),
+			       _( "multi key %s does not exist"),
 			       multi_key.dptr);
 			gripe_corrupt_data();
 		}
