@@ -24,6 +24,7 @@
 #include <stdio.h>
 #include <assert.h>
 #include <errno.h>
+#include <termios.h>
 
 #if defined(STDC_HEADERS)
 #  include <string.h>
@@ -52,7 +53,6 @@ extern pid_t vfork();
 #  include <sys/param.h>
 #endif
 
-#ifdef __profile__
 #  ifndef PATH_MAX
 #    ifdef _POSIX_VERSION
 #      define PATH_MAX _POSIX_PATH_MAX
@@ -64,8 +64,11 @@ extern pid_t vfork();
 #      endif /* MAXPATHLEN */
 #    endif /* _POSIX_VERSION */
 #  endif /* !PATH_MAX */
-#endif /* __profile__ */
 
+	char wd[PATH_MAX];
+	char *cwd = wd;
+
+		
 #ifndef PIPE_BUF
 #  if defined(_POSIX_VERSION) && defined(_POSIX_PIPE_MAX)
 #    define PIPE_MAX _POSIX_PIPE_MAX
@@ -106,12 +109,9 @@ extern int errno;
 #  include "lib/getopt.h"
 #endif /* HAVE_GETOPT_H */
 
-#define NLS_SET	manSet
-#include "nls/nls.h"
-
-#ifdef NLS
-nl_catd catfd;
-#endif /* NLS */
+#include <locale.h>
+#include <libintl.h>
+#define _(String) gettext (String)
 
 #include "manconfig.h"
 #include "libdb/mydbm.h"
@@ -142,12 +142,24 @@ nl_catd catfd;
 #  define STDERR_FILENO 2
 #endif
 
+char * lang;
+struct lt {
+	char *	lang;
+	char *	device;
+	char *	charset;
+} lang_table[] =	{
+	/* LANG		roff_device	LESSCHARSET */
+	{ "ja"		, "nippon"	, "ja"		},
+	{ "*"		, "latin1"	, "latin1"	},
+	{ 0		, 0		, 0		} };
+
 /* external formatter programs, one for use without -t, and one with -t */
 #define NFMT_PROG "./mandb_nfmt"
 #define TFMT_PROG "./mandb_tfmt"
 #undef ALT_EXT_FORMAT	/* allow external formatters located in cat hierarchy */
 
 static int global_manpath = -1;		/* global or user manual page hierarchy? */
+static int skip;		/* page exists but has been skipped */
 
 #if defined _AIX || defined __sgi
 char **global_argv;
@@ -166,7 +178,6 @@ char **global_argv;
 static pid_t *fd2pid = NULL;	/* map fd to pid, unused entries are zero */
 static int max_fd = -1;		/* max fd seen, fd2pid has max_fd+1 elements */
 
-
 FILE *
 popen (const char *cmd, const char *type)
 {
@@ -312,18 +323,45 @@ static FILE *checked_popen (char *command, char *type)
 #endif /* SECURE_MAN_UID */
 
 	if (stream == NULL)
-		error (0, 0,
-		       CATGETS(man_POPEN, "can't popen"));
-
+		error (0, 0, _( "can't popen")); 
 	return stream;
 }
 #endif /* MAN_CATS */
 
+char * lang_dir( char * filename)
+{
+	char *ld;	/* the lang dir: point to static data */
+	char *fm;	/* the first "/man/" dir */
+	char *sm;	/* the second "/man?/" dir */
+
+	ld = "";
+	if ( ! filename ) 
+		return ld;
+
+	if ( ! (fm = strstr( filename, "/man/")) )
+		return ld;
+	if ( ! (sm = strstr( 3+fm, "/man")) )
+		return ld;
+	if ( sm == 4+fm )
+		return ld;
+	if ( sm[5] != '/' )
+		return ld;
+	if ( ! strchr( "123456789lno", sm[4]) )
+		return ld;
+	/* found a lang dir */
+	fm += 5;
+	if ( ! (sm = strchr( fm, '/')) )
+		return ld;
+	ld = xstrdup ( fm);
+	ld[sm-fm] = '\0';
+	if (debug)
+		fprintf (stderr, "found lang dir element %s\n", ld);
+	return ld;
+}
+
 static __inline__ void gripe_system (char *command, int status)
 {
-	error (CHILD_FAIL, 0,
-	       CATGETS(man_SYSTEM, "command exited with status %d: %s"),
-	       status, command);
+	error (CHILD_FAIL, 0, _( "command exited with status %d: %s"), status, command);
 }
 
 
@@ -417,16 +455,23 @@ static const struct option long_options[] =
 
     {"troff", no_argument, 		0, 't'},
     {"troff-device", optional_argument,	0, 'T'},
-#  ifdef TROFF_IS_GROFF
+# ifdef TROFF_IS_GROFF
     {"ditroff", no_argument, 		0, 'Z'},
+    {"gxditview", no_argument, 		0, 'X'},
+#  ifdef ENABLE_HTML
+    {"html", optional_argument,		0, 'H'},
 #  endif
+# endif
     {0, 0, 0, 0}
 };
 
-static const char args[] = "7DlM:P:S:adfhkVum:p:tT:we:L:Zcr:";
+static const char args[] = "7DlM:P:S:adfhH::kVum:p:tT::we:L:Zcr:X";
 
 #  ifdef TROFF_IS_GROFF
 static int ditroff;
+static int htmlout;
+static char *html_pager;
+static int gxditview;
 #  endif /* TROFF_IS_GROFF */
 
 #else /* !HAS_TROFF */
@@ -450,18 +495,18 @@ static void usage (int status)
 #endif /* HAS_TROFF */
 	
 #ifdef HAS_TROFF
-	printf(CATGETS(man_USAGE_TROFF, 
+	printf(_( 
 	       "usage: %s [-c|-f|-k|-w|-tZT device] [-adlhu7V] [-Mpath] [-Ppager] [-Slist]\n"
 	       "           [-msystem] [-pstring] [-Llocale] [-eextension] [section] page ...\n"),
                program_name);
 #else
-	printf(CATGETS(man_USAGE_NOTROFF, 
+	printf(_( 
 	       "usage: %s [-c|-f|-k|-w] [-adlhu7V] [-Mpath] [-Ppager] [-Slist] [-msystem]\n"
                "           [-pstring] [-Llocale] [-eextension] [section] page ...\n"),
                program_name);
 #endif
 
-	puts(CATGETS(man_OPT1, 
+	puts(_( 
 	       "-a, --all                   find all matching manual pages.\n"
                "-d, --debug                 emit debugging messages.\n"
                "-e, --extension             limit search to extension type `extension'.\n"
@@ -474,17 +519,19 @@ static void usage (int status)
                "-c, --catman                used by catman to reformat out of date cat pages.\n"
                "-7, --ascii                 display ASCII translation of certain latin1 chars."));
 #ifdef HAS_TROFF
-	printf(CATGETS(man_OPT2,
+	printf(_(
 	       "-t, --troff                 use %s to format pages.\n"
 	       "-T, --troff-device device   use %s with selected device.\n"),
 	       formatter, formatter);
-#  ifdef TROFF_IS_GROFF
-	puts(CATGETS(man_OPTDIT,
-	       "-Z, --ditroff               use groff and force it to produce ditroff."));
-#  endif /* TROFF_IS_GROFF */
+# ifdef TROFF_IS_GROFF
+#  ifdef ENABLE_HTML
+	puts(_("-H, --html                  use lynx or argument to display html output.\n"));
+#  endif
+	puts(_("-Z, --ditroff               use groff and force it to produce ditroff."));
+# endif /* TROFF_IS_GROFF */
 #endif /* HAS_TROFF */
 
-	puts(CATGETS(man_OPT3,
+	puts(_(
 	       "-D, --default               reset all options to their default values.\n"
                "-M, --manpath path          set search path for manual pages to `path'.\n"
                "-P, --pager pager           use program `pager' to display output.\n"
@@ -494,30 +541,54 @@ static void usage (int status)
 	       "-p, --preprocessor string   string indicates which preprocessors to run."));
 	puts(  "                             e - [n]eqn   p - pic    t - tbl\n"
                "                             g - grap     r - refer  v - vgrind");
-        puts(CATGETS(man_OPT4,
+        puts(_(
                "-V, --version               show version.\n"
                "-h, --help                  show this usage message."));
 
-               exit (status);
+	       exit (status);
 }
 
+/*
+ * changed these messages from stdout to stderr,
+ * (Fabrizio Polacco) Fri, 14 Feb 1997 01:30:07 +0200
+ */
 static void gripe_no_name (char *section)
 {
 	if (section)
-		printf (CATGETS(man_NO_NAME_S, 
-			"What manual page do you want from section %s?\n"), section);
+		fprintf (stderr, _( "What manual page do you want from section %s?\n"), section);
 	else
-		puts (CATGETS(man_NO_NAME, "What manual page do you want?"));
+		fputs ( _( "What manual page do you want?\n"), stderr);
 
 	exit (FAIL);
 }
 
+static struct termios tms;
+static int tms_set=0;
+
+void set_term()
+{
+	if ( tms_set ) tcsetattr( 0, TCSANOW, &tms);
+}
+
+void get_term()
+{
+	if ( isatty( 1) ) {
+		if (debug) fprintf(stderr, "is a tty\n");
+		tcgetattr( 0, &tms);
+		if ( ! tms_set++ )
+			atexit( set_term);
+	}
+}
+
+/*
+ * changed these messages from stdout to stderr,
+ * (Fabrizio Polacco) Fri, 14 Feb 1997 01:30:07 +0200
+ */
 static __inline__ void gripe_no_man(char *name, char *sec)
 {
-	if (troff)
-		printf (CATGETS(man_NO_SRC_MAN, "No source manual entry for %s"), 
-			name);
-	else
+	if (troff) {
+		fprintf ( stderr, _( "No source manual entry for %s"), name);
+	} else {
 
 /* On AIX and IRIX, fall back to the vendor supplied browser */
 #if defined _AIX || defined __sgi
@@ -525,13 +596,14 @@ static __inline__ void gripe_no_man(char *name, char *sec)
 		putenv("MANPATH=");  /* reset the MANPATH variable */
 		execv("/usr/bin/man", global_argv);
 #else
-		printf (CATGETS(man_NO_MAN, "No manual entry for %s"), name);
+		fprintf ( stderr, _( "No manual entry for %s"), name);
 #endif
+	}
 			 
 	if (sec)
-		printf (CATGETS(man_SEC, " in section %s\n"), sec);
+		fprintf ( stderr, _( " in section %s\n"), sec);
 	else
-		putchar ('\n');
+		putc ('\n', stderr);
 }
 
 /* fire up the appropriate external program */
@@ -548,11 +620,11 @@ static __inline__ char **manopt_to_env(int *argc)
 {
 	char *manopt, *opt_start, **argv;
 	
-	opt_start = manopt = getenv("MANOPT");
+	manopt = getenv("MANOPT");
 	if (manopt == NULL || *manopt == '\0')
 		return NULL;
 
-	manopt = xstrdup(manopt);
+	opt_start = manopt = xstrdup(manopt);
 
 	/* allocate space for the program name */
 	*argc = 0;
@@ -569,7 +641,7 @@ static __inline__ char **manopt_to_env(int *argc)
 					*manopt = '\0';
 					argv = (char **) 
 					       xrealloc (argv, (*argc + 3) * 
-						         sizeof (char *));
+							 sizeof (char *));
 					argv[(*argc)++] = opt_start;
 				}
 				while (isspace(*(manopt + 1)))
@@ -657,6 +729,40 @@ static int need_to_rerun(void)
 }
 #endif /* MAN_DB_UPDATES */
 
+
+/* man issued with `-l' option */
+int local_man_loop (char *argv)
+{
+	int exit_status = OK;
+	int local_mf = local_man_file;
+	char *dir;
+
+	local_man_file = 1;
+	if (strcmp (argv, "-") == 0)
+		display (NULL, "", NULL, "(stdin)");
+	else {
+#ifdef COMP_SRC
+		/* See if we need to decompress the file(s) first */
+		struct compression *comp;
+
+		if ( (comp = comp_info(argv)) )
+			(void) decompress(argv, comp);
+#endif /* COMP_SRC */
+		lang = lang_dir (argv);
+		if (!display ((cwd[0]?cwd:NULL), argv, NULL, basename(argv))) {
+			if ( local_mf )
+				error (0, errno, "%s", argv);
+			exit_status = NOT_FOUND;
+		}
+
+#ifdef COMP_SRC
+		remove_ztemp();
+#endif /* COMP_SRC */
+	}
+	local_man_file = local_mf;
+	return (exit_status);
+}
+
 int main (int argc, char *argv[])
 {
 	int argc_env, status = 0, exit_status = OK;
@@ -664,18 +770,22 @@ int main (int argc, char *argv[])
 	char *internal_locale;
 	char *nextarg;
 	extern int optind;
+	void (int_handler)( int);
 
 
-#ifdef __profile__
-	char wd[PATH_MAX];
-	char *cwd = wd;
-#endif /* __profile__ */
-
-                
+	umask(022);
 	/* initialise the locale */
-#ifdef HAVE_SETLOCALE
-	internal_locale = NLS_INIT;
-#endif /* HAVE_SETLOCALE */
+	internal_locale = setlocale( LC_MESSAGES, "");
+	bindtextdomain (PACKAGE, LOCALEDIR);
+	textdomain (PACKAGE);
+
+	if (internal_locale != NULL
+	|| (internal_locale = getenv ("LC_ALL"))
+	|| (internal_locale = getenv ("LC_MESSAGES"))
+	|| (internal_locale = getenv ("LANG")) )
+		internal_locale = xstrdup (internal_locale);
+	else
+		internal_locale = "C";
 
 /* export argv, it might be needed when invoking the vendor supplied browser */
 #if defined _AIX || defined __sgi
@@ -684,6 +794,19 @@ int main (int argc, char *argv[])
 
 	program_name = xstrdup(basename((argv[0])));
 
+	{ /* opens base streams in case someone like "info" closed them */
+		struct stat buf;
+		if ( (fileno(stdin) < 0)
+		|| ((fstat( fileno(stdin), &buf) < 0) && (errno == EBADF)) ) 
+			freopen( "/dev/null", "r", stdin);
+		if ( (fileno(stdout) < 0)
+		|| ((fstat( fileno(stdout), &buf) < 0) && (errno == EBADF)) ) 
+			freopen( "/dev/null", "w", stdout);
+		if ( (fileno(stderr) < 0)
+		|| ((fstat( fileno(stderr), &buf) < 0) && (errno == EBADF)) ) 
+			freopen( "/dev/null", "w", stderr);
+	}
+
 /* if the user happens to be root and man is installed setuid, forget all
  * of the priv dropping etc. Just make sure that any cat files produced
  * are owned by MAN_OWNER by doing explicit chown() calls */
@@ -700,7 +823,6 @@ int main (int argc, char *argv[])
 	drop_effective_privs();
 #endif
 
-#ifdef __profile__
 	/* This will enable us to do some profiling and know
 	where gmon.out will end up. Must chdir(cwd) before we return */
 #  ifdef HAVE_GETCWD
@@ -709,7 +831,6 @@ int main (int argc, char *argv[])
 	if (!getwd(cwd))
 #  endif
 		cwd[0] = '\0';
-#endif /* __profile__ */
 
 	/* First of all, find out if $MANOPT is set. If so, put it in 
 	   *argv[] format for getopt to play with. */
@@ -725,18 +846,29 @@ int main (int argc, char *argv[])
 	if (external)
 		do_extern (argv);
 
+	get_term(); /* stores terminal settings */
+	if (debug) fprintf(stderr, "real user = %d; effective user = %d\n", ruid, euid);
+
 #ifdef HAVE_SETLOCALE
 	/* close this locale and reinitialise incase a new locale was 
 	   issued as an argument or in $MANOPT */
 	if (locale) {
-		internal_locale = setlocale(LC_MESSAGES, locale);
-#  ifdef NLS
-		NLS_CLOSE;
-		catfd = catopen("man_db", MCLoadBySet);
-#  endif /* NLS */
+		if ((internal_locale = setlocale(LC_MESSAGES, locale)) == NULL)
+			internal_locale = locale;
+
+		internal_locale = xstrdup (internal_locale);
+if (debug) fprintf(stderr, "main(): locale= %s,internal_locale= %s\n", locale, internal_locale);
+		if (internal_locale) {
+			extern int  _nl_msg_cat_cntr;
+			setenv ("LANGUAGE", internal_locale, 1);
+			++_nl_msg_cat_cntr;
+		}
 	}
+
 #endif /* HAVE_SETLOCALE */
 	
+	if ( htmlout )
+		pager = (html_pager == NULL ?  WEB_BROWSER : html_pager);
 	if (pager == NULL)
 		if ((pager = getenv ("PAGER")) == NULL)
 			pager = PAGER;
@@ -747,10 +879,10 @@ int main (int argc, char *argv[])
 #ifdef LESS_PROMPT
 		prompt_string = LESS_PROMPT;
 #else
-		prompt_string = CATGETS(man_LESS_PROMPT, 
-					" Manual page $MAN_PN "
-					"?ltline %lt?L/%L.:byte %bB?s/%s..?e (END):"
-					"?pB %pB\\\\%..");
+		prompt_string = _( 
+				" Manual page $MAN_PN "
+				"?ltline %lt?L/%L.:byte %bB?s/%s..?e (END):"
+				"?pB %pB\\\\%..");
 #endif
 
 	if (debug)
@@ -759,33 +891,12 @@ int main (int argc, char *argv[])
 	if (optind == argc)
 		gripe_no_name (NULL);
 
-	putenv("LESSCHARSET=latin1");
+	signal( SIGINT, int_handler);
 
 	/* man issued with `-l' option */
 	if (local_man_file) {
-		for (;  optind < argc;  ++optind) {
-			if (strcmp (argv[optind], "-") == 0)
-				display (NULL, "", NULL, "(stdin)");
-			else {
-				/* See if we need to decompress the file(s)
-				   first */
-#ifdef COMP_SRC
-				struct compression *comp;
-
-				if ( (comp = comp_info(argv[optind])) )
-					(void) decompress(argv[optind], comp);
-#endif /* COMP_SRC */
-				if (!display (NULL, argv[optind], NULL, 
-					     basename(argv[optind]))) {
-					error (0, errno, argv[optind]);
-					exit_status = NOT_FOUND;
-				}
-
-#ifdef COMP_SRC
-				remove_ztemp();
-#endif /* COMP_SRC */
-			}
-		}
+		for (;  optind < argc;  ++optind)
+			exit_status = local_man_loop( argv[optind]);
 		exit (exit_status);
 	}
 
@@ -807,7 +918,6 @@ int main (int argc, char *argv[])
 
 	if (optind == argc - 1) {
 		tmp = is_section ( argv[optind] );
-
 		if (tmp)
 			gripe_no_name (tmp);
 	}
@@ -825,28 +935,30 @@ int main (int argc, char *argv[])
      		 * See if this argument is a valid section name.  If not,
       		 * is_section returns NULL.
       		 */
-
-		tmp = is_section (nextarg);
-
-		if (tmp) {
-			section = tmp;
-
-			if (debug)
-				fprintf (stderr, "\nsection: %s\n", section);
-
-			continue;
+		if (optind < argc) {
+			if (tmp = is_section (nextarg)) {
+				section = tmp;
+				if (debug)	fprintf (stderr, "\nsection: %s\n", section);
+				continue;
+			}
 		}
 
 		/* this is where we actually start looking for the man page */
-
+		skip = 0;
 		status = man (nextarg);
 
 		/* clean out the memory cache for each man page */
 		free_hashtab();
 
 		if (!status && !catman) {
-			gripe_no_man(nextarg, section);
-			exit_status = NOT_FOUND;
+			if ( !skip ) {
+				if ( strchr( nextarg, '/') )
+					exit_status = local_man_loop( nextarg);
+				else
+					exit_status = NOT_FOUND;
+				if ( exit_status == NOT_FOUND)
+					gripe_no_man(nextarg, section);
+			}
 		} else {
 			if (debug)
 				fprintf(stderr,
@@ -866,11 +978,9 @@ int main (int argc, char *argv[])
 		chkr_garbage_detector();
 	}
 
-#ifdef __profile__
 	/* For profiling */
 	if (cwd[0])
 		chdir(cwd);
-#endif /* __profile__ */
 
 	exit (exit_status);
 }
@@ -899,7 +1009,7 @@ static void man_getopt (int argc, char *argv[])
 				pager = optarg;
 				break;
 		    	case 'S':
-				colon_sep_section_list = optarg;
+				if (*optarg) colon_sep_section_list = optarg;
 				break;
 			case 'V':
 				ver();
@@ -945,14 +1055,24 @@ static void man_getopt (int argc, char *argv[])
 
 			case 'T':
 				/* @@@ traditional nroff knows -T,
-                                   troff does not (gets ignored).
-                                   All incarnations of groff know it.
-                                   Why does -T imply -t? */
+				   troff does not (gets ignored).
+				   All incarnations of groff know it.
+				   Why does -T imply -t? */
 				/* as this is an optional argument */
-				if (optarg)
-					roff_device = optarg;
+				roff_device = (optarg ? optarg : "ps");
 				troff = 1;
 				break;
+			case 'X':
+#ifdef TROFF_IS_GROFF
+				gxditview = troff = 1;
+#endif/* TROFF_IS_GROFF */
+				break;
+			case 'H':
+#ifdef TROFF_IS_GROFF
+				html_pager = (optarg ? optarg : 0);
+				htmlout = 1;
+#endif/* TROFF_IS_GROFF */
+				break;
 			case 'Z':
 #ifdef TROFF_IS_GROFF
 				ditroff = 1;
@@ -969,7 +1089,7 @@ static void man_getopt (int argc, char *argv[])
 			case 'D':
 		    		/* discard all preset options */
 		    		local_man_file = findall = update = catman = debug
-		    			 = troff
+		    			 = troff = gxditview
 		    			 = print_where = ascii = 0;
 #ifdef TROFF_IS_GROFF
 				ditroff = 0;
@@ -997,31 +1117,26 @@ static void man_getopt (int argc, char *argv[])
 		       		 apropos ? "-k " : "",
 		       		 catman ? "-c " : "",
 		       		 print_where ? "-w " : "", 
-		       		 CATGETS(man_INCOMP_OPT,
-		       		          ": incompatible options"), NULL));
+		       		 _( ": incompatible options"), NULL));
 		usage(FAIL);
 	}
 }
 
 /*
- * Check to see if the argument is a valid section number.  If the
- * first character of name is a numeral, or the name matches one of
+ * Check to see if the argument is a valid section number. 
+ * If the name matches one of
  * the sections listed in section_list, we'll assume that it's a section.
  * The list of sections in config.h simply allows us to specify oddly
  * named directories like .../man3f.  Yuk.
  */
 static __inline__ char *is_section (char *name)
 {
+	char **vs;
 
-	if ( !(isdigit (name[0]) && !isdigit (name[1])) ) {
-		char **vs;
-	
-		for (vs = section_list; *vs; vs++)
-			if ((strcmp (*vs, name) == 0)) 
-				return name;
-		return NULL;
-	}
-	return name;
+	for (vs = section_list; *vs; vs++)
+		if ((strcmp (*vs, name) == 0)) 
+			return name;
+	return NULL;
 }
 
 /* Snarf pre-processors from file, return (static) string or NULL on failure */
@@ -1050,7 +1165,7 @@ static char *get_preprocessors_from_file (char *file)
 	}
 
 	/* if we couldn't read the first line from file, or we didn't
-           find PP_COOKIE, then directive == NULL */
+	   find PP_COOKIE, then directive == NULL */
 #endif
 	return directive;
 }
@@ -1085,13 +1200,101 @@ static char *get_preprocessors (char *file)
 	return pp_string;
 }
 
+/* This is so that we can store the temp file name used when input is
+ * stdin and remove it on exit.
+ */
+static char *stdin_tmpfile;
+static int stdin_tmpfile_fd;
 
+static void remove_stdintmp(void)
+{
+	if (stdin_tmpfile) {
+		if (stdin_tmpfile_fd >= 0)
+			close(stdin_tmpfile_fd);
+		(void) remove_with_dropped_privs(stdin_tmpfile);
+	}
+	stdin_tmpfile = NULL;
+}
+
+static __inline__ void create_stdintmp(void)
+{
+	/* This is basically copied from create_ztemp() in compression.c;
+	 * it therefore shouldn't introduce any new security holes -- PMM
+	 */
+	int oldmask = umask(022);
+	drop_effective_privs();
+	errno = 0;				/* failing tempname() might fail to set errno */
+	stdin_tmpfile = tempnam(NULL, "sman");
+	if (stdin_tmpfile)
+		stdin_tmpfile_fd = open(stdin_tmpfile, O_WRONLY|O_CREAT|O_TRUNC|O_EXCL, 0644);
+	if (!stdin_tmpfile || (stdin_tmpfile_fd < 0))
+		error (FATAL, errno, _( "can't create a temporary filename"));
+	regain_effective_privs();
+	umask(oldmask);
+	atexit( remove_stdintmp);
+}
+  
 /* Return command (malloced string) to format file to stdout */
 static __inline__ char *make_roff_command (char *dir, char *file)
 {
-	char *pp_string = get_preprocessors (file);
+	char *pp_string;
 	char *fmt_prog;
 	char *command;
+	char *tmpfile;
+
+	if (!*file) {
+		/* file == "": this means we are reading input from stdin.
+		 * Unfortunately we need to read the first line or so to
+		 * determine the preprocessor filename. So we read a bit
+		 * into a temporary file which we give to get_preprocessors
+		 * and then prepend to the rest of stdin by putting cat
+		 * in the beginning of the pipeline we pass to system().
+		 * I can't think of a better way to do this :-<
+		 * [this code seems remarkably hairy for what it does...]
+		 */
+		/* get_preprocessors uses 128 byte buffer, no point using more here */
+#define STDIN_SNARFLEN 128
+		char bf[STDIN_SNARFLEN];
+		int ct = 0;		       /* number of bytes in bf[] */
+		int ctot, r;
+		/* Ignore SIGPIPE; we want to be notified by write returning EPIPE. */
+		RETSIGTYPE (*old_handler)() = signal(SIGPIPE, SIG_IGN);
+
+		/* read data into bf; we don't want to use the stream stuff because
+		 * we don't want data to be lost in the buffers, not read here or
+		 * by the system()-run pipeline.
+		 */
+		do {
+			errno = 0;
+			r = read(STDIN_FILENO, bf+ct, (128-ct));
+			if (r != -1)
+				ct += r;
+			/* stop on EOF (r==0), bf full (ct==128) or error (errno not 0 or EINTR) */
+		} while (r != 0 && ct < STDIN_SNARFLEN && (errno == 0 || errno == EINTR));
+
+		if (errno != 0)
+			error(FATAL, errno, _( "error trying to read from stdin"));
+
+		create_stdintmp();
+		/* write bf to stdin_tmpfile_fd */
+		ctot = ct;
+		do {
+			errno = 0;
+			r = write(stdin_tmpfile_fd, bf+(ctot-ct), ct);
+			if (r != -1)
+				ct -= r;
+		} while (ct > 0 && (errno == 0 || errno == EINTR));
+		if (errno != 0)
+			error(FATAL, errno, _( "error writing to temporary file %s"), stdin_tmpfile);
+
+		close(stdin_tmpfile_fd);
+		stdin_tmpfile_fd = -1;	    /* ensure we don't try to close it again */
+		signal(SIGPIPE, old_handler);
+
+		pp_string = get_preprocessors(stdin_tmpfile);
+	}
+	else
+		pp_string = get_preprocessors(file);
 
 #ifdef ALT_EXT_FORMAT
 	/* Check both external formatter locations */
@@ -1136,17 +1339,49 @@ static __inline__ char *make_roff_command (char *dir, char *file)
 		char *dev;	/* either " -T<mumble>" or "" */
 		int using_tbl = 0;
 
-		if (roff_device)
+		/* load the roff_device value dependent on the language dir in path */
+		if ( ! roff_device && ! troff ) {
+			if (( ! lang ) || ( ! *lang )) {
+				roff_device = "latin1";
+				putenv( "LESSCHARSET=latin1");
+			} else {
+				int j;
+				for ( j=0; lang_table[j].lang; j++ ) {
+					if (( strncmp( lang_table[j].lang, lang
+						, strlen( lang_table[j].lang)) == 0 )
+					||  ( lang_table[j].lang[0] == '*' )) {
+						roff_device = lang_table[j].device;
+						putenv( strappend ( 0
+							,"LESSCHARSET="
+							, lang_table[j].charset
+							, 0));
+						j = sizeof(lang_table)/sizeof(struct lt) - 2;
+					}
+				}
+			}
+		}
+		/* tell grops to guess the page size */
+		if ( roff_device && strcmp( roff_device, "ps") == 0 )
+			roff_device = strappend( NULL, "ps -P-g ", NULL);
+		if (gxditview && roff_device)
+			dev = strappend (NULL, " -X -T", roff_device, NULL);
+		else if (gxditview)
+			dev = strappend (NULL, " -X", NULL);
+		else if (roff_device)
 			dev = strappend (NULL, " -T", roff_device, NULL);
 		else
 			dev = "";
 
 		if (*file)
-			command = strappend (NULL, SOELIM " '", 
-					     file, "'", NULL); 
-		else
-			command = xstrdup (SOELIM);
-
+			command = strappend (NULL, SOELIM " '", file, "'", NULL); 
+		else {
+			/* Reading from stdin: use cat to pick up the part we
+			 * read in to figure out the format pipeline.
+			 * ? is '-' as a cat argument standard?
+			 * If not we could try "(cat tempfile; cat) | SOELIM..."
+			 */
+			command = strappend(NULL, "cat ", stdin_tmpfile, " - | ", SOELIM, NULL);
+		}
 		do {
 			char *filter;
 			int wants_dev = 0; /* filter wants a dev argument */
@@ -1208,10 +1443,7 @@ static __inline__ char *make_roff_command (char *dir, char *file)
 						     NULL);
 			} else {
 				assert (*pp_string); /* didn't fail on roff */
-				error (0, 0,
-				       CATGETS(man_BAD_PP,
-					       "ignoring unknown preprocessor `%c'"),
-				       *pp_string);
+				error (0, 0, _( "ignoring unknown preprocessor `%c'"), *pp_string);
 			      }
 		} while (*pp_string++);
 
@@ -1221,7 +1453,7 @@ static __inline__ char *make_roff_command (char *dir, char *file)
 			command = strappend (command, " | ", COL, NULL);
 #endif /* GNU_NROFF */
 
-		if (roff_device) free (dev);
+		if (roff_device || gxditview) free (dev);
 	} else {
 		/* use external formatter script, it takes arguments
 		   input file, preprocessor string, and (optional)
@@ -1272,7 +1504,7 @@ static __inline__ int do_chown(const char *file)
 	int status;
 	
 	if ((status = chown (file, euid, (uid_t) -1)) )
-		error (0, errno, CATGETS (man_CHOWN, "can't chown %s"), file);
+		error (0, errno, _( "can't chown %s"), file);
 	
 	return status;
 }
@@ -1305,9 +1537,7 @@ static int commit_tmp_cat (char *cat_file, char *tmp_cat, int delete)
 		} else {
 			status = chmod (tmp_cat, CATMODE);
 			if (status)
-				error(0, errno,
-				      CATGETS(man_CHMOD, "can't chmod %s"),
-				      tmp_cat);
+				error(0, errno, _( "can't chmod %s"), tmp_cat);
 		}
 	} else
 		status = 0;
@@ -1322,9 +1552,7 @@ static int commit_tmp_cat (char *cat_file, char *tmp_cat, int delete)
 		} else {
 			status = do_chown(tmp_cat);
 			if (status)
-				error(0, errno,
-				      CATGETS(man_CHOWN, "can't chown %s"),
-				      tmp_cat);
+				error(0, errno, _( "can't chown %s"), tmp_cat);
 		}
 	}
 #endif /* SECURE_MAN_UID && MAN_CATS */
@@ -1336,10 +1564,7 @@ static int commit_tmp_cat (char *cat_file, char *tmp_cat, int delete)
 			status = 0;
 		} else {
 			if ((status = rename (tmp_cat, cat_file)))
-				error (0, errno,
-				       CATGETS (man_RENAME,
-						"can't rename %s to %s"),
-				       tmp_cat, cat_file);
+				error (0, errno, _( "can't rename %s to %s"), tmp_cat, cat_file);
 		}
 	}
 
@@ -1348,10 +1573,7 @@ static int commit_tmp_cat (char *cat_file, char *tmp_cat, int delete)
 			fprintf (stderr, "unlinking temporary cat\n");
 		else
 			if (unlink (tmp_cat))
-				error (0, errno,
-				       CATGETS (man_UNLINK,
-						"can't unlink %s"),
-				       tmp_cat);
+				error (0, errno, _( "can't unlink %s"), tmp_cat);
 	}
 
 	return status;
@@ -1378,7 +1600,7 @@ static __inline__ FILE *open_cat_stream (char *cat_file)
 
 	/* create pipe */
 	if (pipe (pipe_fd)) {
-		error (0, errno, CATGETS (man_PIPE, "can't create pipe"));
+		error (0, errno, _( "can't create pipe"));
 		return NULL;
 	}
 
@@ -1386,7 +1608,7 @@ static __inline__ FILE *open_cat_stream (char *cat_file)
 	fflush (NULL);
 	child = fork ();
 	if (child < 0) {
-		error (0, errno, CATGETS (man_FORK, "can't fork"));
+		error (0, errno, _( "can't fork"));
 		return NULL;
 	} else if (child == 0) {
 		/* compress reading end of pipe to tmp_cat_file */
@@ -1403,19 +1625,17 @@ static __inline__ FILE *open_cat_stream (char *cat_file)
 
 		/* connect standard input to reading end of pipe */
 		if (dup2 (pipe_fd[0], 0) == -1)
-			error (0, errno, CATGETS (man_DUP2, "can't dup2"));
+			error (0, errno, _( "can't dup2"));
 		close (pipe_fd[0]);
 		close (pipe_fd[1]);
 
 		/* set up standard output to write to tmp_cat_file */
 		cat_fd = open (tmp_cat_file, O_WRONLY|O_CREAT|O_TRUNC, CATMODE);
 		if (cat_fd == -1) {
-			error (FATAL, errno,
-			       CATGETS (man_CREATE_CAT, "can't create %s"),
-			       tmp_cat_file);
+			error (FATAL, errno, _( "can't create %s"), tmp_cat_file);
 		}
 		if (dup2 (cat_fd, 1) == -1)
-			error (0, errno, CATGETS (man_DUP2, "can't dup2"));
+			error (0, errno, _( "can't dup2"));
 		close (cat_fd);
 
 		nice (10);
@@ -1434,9 +1654,7 @@ static __inline__ FILE *open_cat_stream (char *cat_file)
 				argv[n] = strtok (NULL, " \t\n");
 			}
 			execve (path, argv, envp);
-			error (FATAL, errno,
-			       CATGETS (man_EXEC, "can't exec %s"),
-			       path);
+			error (FATAL, errno, _( "can't exec %s"), path);
 		}
 	}
 
@@ -1473,9 +1691,7 @@ static __inline__ int close_cat_stream (FILE *cat_stream, char *cat_file, int de
 
 		while (waitpid (cat_comp_pid, &comp_status, 0) == -1) {
 			if (errno != EINTR) {
-				error (0, errno,
-				       CATGETS (man_WAITPID, "waiting for pid %u"),
-				       cat_comp_pid);
+				error (0, errno, _( "waiting for pid %u"), cat_comp_pid);
 				comp_status = -1;
 				break;
 			}
@@ -1525,9 +1741,7 @@ static int format_display_and_save (char *roff_cmd, char *disp_cmd, char *cat_fi
 						    out);
 					outed += n;
 					if (!n && saving)
-						fprintf (stderr,
-							 CATGETS(man_STILL_SAVING, 
-							"Still saving the page, please wait...\n"));
+						fprintf (stderr, _( "Still saving the page, please wait...\n"));
 					outing = n; /* stop outing on error */
 				}
 
@@ -1555,6 +1769,8 @@ static int format_display_and_save (char *roff_cmd, char *disp_cmd, char *cat_fi
  * optionally chdir to dir, if necessary update cat_file from man_file
  * and display it.  if man_file is NULL cat_file is a stray cat.  If
  * !save_cat or cat_file is NULL we must not save the formatted cat.
+ * If man_file is "" this is a special case -- we expect the man page
+ * on standard input.
  */
 static int display (char *dir, char *man_file, char *cat_file, char *title)
 {
@@ -1568,9 +1784,7 @@ static int display (char *dir, char *man_file, char *cat_file, char *title)
 			fprintf (stderr, "chdir %s\n", dir);
 
 		if (chdir (dir)) {
-			error (0, errno,
-			       CATGETS (man_CD, "can't chdir to %s"),
-			       dir);
+			error (0, errno, _( "can't chdir to %s"), dir);
 			return 0;
 		}
 	}
@@ -1663,7 +1877,15 @@ static int display (char *dir, char *man_file, char *cat_file, char *title)
 			free (cat_dir);
 		}
 
-		found = !access (format ? man_file : cat_file, R_OK);
+		/* if we're trying to read stdin via '-l -' then man_file will be ""
+		* which access() obviously barfs on, but all is well because the
+		* roff_cmd will have been created to expect input via stdin.
+		* So we special-case this to avoid the bogus access() check.
+		*/
+		if (format == 1 && *man_file == '\0')
+			found = 1;
+		else
+			found = !access (format ? man_file : cat_file, R_OK);
 
 		if (debug)
 			fprintf(stderr, "format: %d, save_cat: %d, found: %d\n",
@@ -1679,10 +1901,7 @@ static int display (char *dir, char *man_file, char *cat_file, char *title)
 			} else if (catman) {
 				if (format) {
 					if (!save_cat) {
-						error (0, 0,
-						       CATGETS (man_NO_CAT, 
-						       		"\ncannot write to %s in catman mode"),
-						       cat_file);
+						error (0, 0, _( "\ncannot write to %s in catman mode"), cat_file);
 					} else {
 						char *tmpcat = tmp_cat_filename (cat_file);
 						cmd = strappend (NULL, roff_cmd,
@@ -1722,10 +1941,7 @@ static int display (char *dir, char *man_file, char *cat_file, char *title)
 						return 0;
 				}
 
-				fprintf (stderr,
-					CATGETS(man_REFORMAT, 
-					"Reformatting %s, please wait...\n"),
-					title);
+				fprintf (stderr, _( "Reformatting %s, please wait...\n"), title);
 				disp_cmd = make_display_command (NULL, title);
 	
 #ifdef MAN_CATS
@@ -1887,6 +2103,7 @@ static int try_section (char *path, char *sec, char *name)
 
 			if (debug)
 				fprintf (stderr, "found ultimate source file %s\n", man_file);
+			lang = lang_dir (man_file);
 
 			cat_file = find_cat_file (path, man_file, sec);
 			found += display (path, man_file, cat_file, title);
@@ -1967,9 +2184,7 @@ static int try_db_section (char *orig_name, char *path, struct mandata *in)
 				name = orig_name;
 		} 
 		  else
-			error (0, errno,
-			       CATGETS(man_UPDATE_DB, "can't update index cache %s"),
-			       database);
+			error (0, errno, _( "can't update index cache %s"), database);
 	}
 #endif /* MAN_DB_UPDATES */
 
@@ -1996,6 +2211,7 @@ static int try_db_section (char *orig_name, char *path, struct mandata *in)
 
 			if (debug)
 				fprintf (stderr, "found ultimate source file %s\n", man_file);
+			lang = lang_dir (man_file);
 
 			cat_file = find_cat_file (path, man_file, in->ext);
 			found += display (path, man_file, cat_file, title);
@@ -2070,8 +2286,10 @@ static int exist_check(char *name, char *manpath, struct mandata *loc)
 	exists = try_db_section(name, manpath, loc);
 			
 #ifdef MAN_DB_UPDATES
-	if (exists == 0)
+	if ( !exists && !skip ) {
+		if (debug) fprintf(stderr, "dbdelete_wrapper =%s =%s\n", name, loc);
 		dbdelete_wrapper(name, loc);
+	}
 #endif /* MAN_DB_UPDATES */
 
 	return exists;
@@ -2123,7 +2341,7 @@ static int try_db(char *manpath, char *sec, char *name)
 			/* create_db should really return EOF on failure. */
 			if (create_db (manpath) == 0) {
 				data = infoalloc();
-				data->next = data->addr = NULL;
+				data->next = NULL; data->addr = NULL;
 				(void) install_db_ptr(manpath, data);
 				return -1;
 			}
@@ -2256,7 +2474,8 @@ static int man (char *name)
 	} else {
 		char **sp;
 		
-		for (sp = section_list; *sp; sp++) {
+		/* sections in manpath are only 1 char */
+		for (sp = section_list; *sp && !sp[0][1]; sp++) {
 			char **mp;
 			
 			for (mp = manpathlist; *mp; mp++) {
@@ -2314,18 +2533,23 @@ static __inline__ char **get_section_list (void)
  */
 static __inline__ int do_prompt (char *name)
 {
+	int ch;
+
+	skip = 0;
 	if (isatty (STDOUT_FILENO) && isatty (STDIN_FILENO)) {
-		fprintf (stderr, CATGETS(man_NEXT, 
+		fprintf (stderr, _( 
 		         "--Man-- next: %s "
 		         "[ view (return) | skip (Ctrl-D) | quit (Ctrl-C) ]\n"), 
 		         name);
 		fflush (stderr);
 
 		do {
-			switch( getchar() ) {
+			ch=getchar();
+			switch( ch ) {
 				case '\n':
 					return 0;
 				case EOF:
+					skip = 1;
 					return 1;
 				default:
 					break;
@@ -2334,3 +2558,10 @@ static __inline__ int do_prompt (char *name)
 	}
 	return 0;
 }
+
+void (int_handler)( int signo)
+{
+	if (debug)
+		fprintf (stderr, "\ninterrupt signal %d handler\n", signo);
+	exit( INTERRUPTED);
+}
