@@ -79,12 +79,13 @@ extern int errno;
 #  endif /* _POSIX_VERSION */
 #endif /* !PATH_MAX */
 
-#define NLS_SET	manpSet
-#include "nls/nls.h"
+#include <libintl.h>
+#define _(String) gettext (String)
 
 #include "manconfig.h"
 #include "lib/error.h"
 #include "manp.h"
+#include "security.h"
 
 typedef struct {
 	char mandir[PATH_MAX];
@@ -100,7 +101,6 @@ DIRLIST list[MAXDIRS];
 char *tmplist[MAXDIRS];
 char *manpathlist[MAXDIRS];
 
-static __inline__ int get_dirlist(void);
 static __inline__ char *get_manpath(char *path);
 static __inline__ char *has_subdirs(char *p);
 static __inline__ char *fsstnd(char *path);
@@ -112,27 +112,22 @@ static char **add_dir_to_path_list(char **mp, char *p);
 static __inline__ void gripe_reading_mp_config (char *file)
 {
 	if (!quiet)
-		error (FAIL, 0,
-		       CATGETS(manp_PARSE_CONF,
-			       "can't make sense of the manpath configuration file %s"), 
-		       file);
+		error (FAIL, 0, _( "can't make sense of the manpath configuration file %s"), file);
 }
 
 static __inline__ void gripe_stat_file(char *file)
 {
 	if (!quiet)
-		error (0, errno,
-		       CATGETS(manp_STAT, "warning: %s"), file);
+		error (0, errno, _( "warning: %s"), file);
 }
 
 static __inline__ void gripe_not_directory(char *dir)
 {
 	if (!quiet)
-		error (0, 0,
-		       CATGETS(manp_NOT_DIR, "warning: %s isn't a directory"),
-		       dir);
+		error (0, 0, _( "warning: %s isn't a directory"), dir);
 }
 
+
 /* accept a manpath list, separated with ':', return the associated 
    catpath list */
 char *cat_manpath(char *manp)
@@ -173,51 +168,88 @@ char *cat_manpath(char *manp)
 	return ++catp;
 }		
 
+static char *
+check_and_give (const char *path, const char *locale)
+{
+	char *result = NULL, *testdir;
+	int test;
+
+	testdir = strappend (NULL, path, "/", locale, NULL);
+	test = is_directory (testdir);
+
+	if ( test == 1 ) {
+		if (debug)
+			fprintf (stderr, "check_and_give(): adding %s\n", testdir);
+
+		result = xstrdup (testdir);	/* I do not like side effects */
+	} else if ( test == 0 ) 
+		gripe_not_directory (testdir);
+	
+	free (testdir);
+
+	return result;
+}
+
+static char *
+add_to_manpath (char *manpath, const char *path)
+{
+	char *result;
+
+	if (manpath == NULL)
+		result = xstrdup (path);
+	else
+		result = strappend (manpath, ":", path, NULL);
+
+	return result;
+}
+
+
 char *add_nls_manpath (char *manpathlist, const char *locale)
 {
 #ifdef HAVE_SETLOCALE
 	char *manpath = NULL;
 	char *path;
+	char *temp_locale;
+	char *omanpathlist = strdup(manpathlist);
 
-	if (locale == NULL || *locale == '\0')
+	if (debug)
+		fprintf (stderr, "add_nls_manpath(): processing %s\n", manpathlist);
+
+	if (locale == NULL || *locale == '\0' || *locale == 'C')
 		return manpathlist;
 
-	for (path = strtok(manpathlist, ":"); path; 
-	     path = strtok(NULL, ":") ) {
-	     
-		char *testdir;
-		int test;
-	     	
-		testdir = strappend (NULL, path, "/", locale, NULL);
-	     	test = is_directory(testdir);
+	temp_locale = xstrdup( locale);
 
-	     	if ( test == 1 ) {
-	     		if (debug)
-	     			fprintf(stderr, 
-	     				"add_nls_manpath(): adding %s\n",
-	     				testdir);
-	     		if (!manpath)
-	     			manpath = xstrdup (testdir);
-	     		else
-	     			manpath = strappend (manpath, ":", testdir,
-	     					     NULL);
-	     	} else if ( test == 0 ) 
-	     		gripe_not_directory(testdir);
-	     	
-	     	free(testdir);
+	for (path = strtok( manpathlist, ":"); path;
+	     path = strtok( NULL, ":") ) {
 
-		if (!manpath)
-			manpath = xstrdup (path);
-		else
-			manpath = strappend (manpath, ":", path, NULL);
+		static char locale_delims[] = "@,._";
+		char *delim, *tempo;
+		char *testpath;
+
+		strcpy( temp_locale, locale);
+
+		if ( (testpath = check_and_give( path, temp_locale)) ) {
+			manpath = add_to_manpath (manpath, testpath);
+			free (testpath);
+		}
+		for (delim = locale_delims ; *delim != '\0' ; ++delim)
+			if ( (tempo = strchr (temp_locale, *delim)) ) {
+				*tempo = '\0';	/* Strip out the rest of the line */
+				if ( (testpath = check_and_give (path, temp_locale)) ) {
+					manpath = add_to_manpath (manpath, testpath);
+					free (testpath);
+				}
+			}
+	}
+	/* After doing all the locale stuff we add the manpath to the *END*
+	 * so the locale dirs are checked first on each section */
+	manpath = add_to_manpath(manpath,omanpathlist);
+	free (omanpathlist);
+
+	free (manpathlist);
+	return manpath;
 
-	 }
-	 free(manpathlist);
-#if 0
-	 if (debug)
-	 	fprintf(stderr, "path inc nls = %s\n", manpath);
-#endif
-	 return manpath;
 #else /* !HAVE_SETLOCALE */
 	return manpathlist;
 #endif /* HAVE_SETLOCALE */
@@ -233,7 +265,7 @@ char *add_system_manpath (char *systems, char *manpathlist)
 
 	if (systems == NULL)
 		if ((systems = getenv ("SYSTEM")) == NULL)
-			return manpathlist;
+			systems = "";
 	if (*systems == '\0')
 		return manpathlist;
 
@@ -285,9 +317,11 @@ char *add_system_manpath (char *systems, char *manpathlist)
 						*manpath = '\0';
 					strcat (manpath, newdir);
 					manpathsize += newdirlen;
-				} else if (debug) {
+				} else {
+					if (debug) {
 					fputs ("can't stat ", stderr);
 					perror (newdir);
+					}
 				}
 				free (newdir);
 			}
@@ -305,9 +339,17 @@ char *add_system_manpath (char *systems, char *manpathlist)
 			manpathsize += pathlen;
 		}
 	}
-	if (!manpath)
-		error(FAIL, 0, CATGETS(manp_NULL_MANPATH,
-				    "internal manpath equates to NULL"));
+	/*
+	 * Thu, 21 Nov 1996 22:24:19 +0200 fpolacco@debian.org
+	 * bug#5534 (man fails if env var SYSTEM is defined)
+	 * with error [man: internal manpath equates to NULL]
+	 * the reason: is_directory (newdir); returns -1
+	 */
+	if (!manpath) {
+		if (debug) fprintf(stderr, "add_system_manpath(): %s\n",
+				    "internal manpath equates to NULL");
+		return manpathlist;
+	}
 	return manpath;
 }
 
@@ -327,41 +369,17 @@ char *add_system_manpath (char *systems, char *manpathlist)
  * $HOME/bin in his path and the directory $HOME/bin/man exists -- the
  * directory $HOME/bin/man will be added to the manpath.
  */
-char *manpath (char *systems)
+char *guess_manpath (char *systems)
 {
 	int len;
-	char *manpathlist;
 	char *path;
 
-	/* need to read config file even if MANPATH set, for mandb(8) */
-	
-	if (get_dirlist ())
-		gripe_reading_mp_config (CONFIG_FILE);
-
-	if ((manpathlist = getenv ("MANPATH")) ) {
-		/*
-    		 * This must be it.
-    		 */
-		if (!quiet)
-			error (0, 0,
-			       CATGETS(manp_MANPATH,
-				       "warning: $MANPATH set, ignoring %s"),
-			       CONFIG_FILE);
-		/* xstrdup() required below so that we don't hack up the
-		   environment */
-
-		manpathlist = add_system_manpath (systems, xstrdup (manpathlist));
-
-		return manpathlist;
-	}
-
 	if ((path = getenv ("PATH")) == NULL) {
 		/*
 		 * Things aren't going to work well, but hey...
 		 */
 		if (!quiet)
-			error (0, 0, CATGETS(manp_NO_PATH,
-					     "warning: $PATH not set"));
+			error (0, 0, _( "warning: $PATH not set"));
 		
 		path = add_system_manpath (systems, def_path (MANDATORY));
 
@@ -372,9 +390,7 @@ char *manpath (char *systems)
 			 * Things aren't going to work well here either...
 			 */
 			if (!quiet)
-				error (0, 0,
-				       CATGETS(manp_MT_PATH,
-					       "warning: empty $PATH"));
+				error (0, 0, _( "warning: empty $PATH"));
 			
 			path = add_system_manpath (systems, def_path (MANDATORY));
 
@@ -387,27 +403,139 @@ char *manpath (char *systems)
 	}
 }
 
+char *manpath (char *systems)
+{
+	char *manpathlist;
+
+	/* need to read config file even if MANPATH set, for mandb(8) */
+	if (get_dirlist ())
+		gripe_reading_mp_config (CONFIG_FILE);
+
+	manpathlist = getenv ("MANPATH");
+	if ( manpathlist && *manpathlist ) {
+		char *pos;
+		/*
+    		 * This must be it.
+    		 */
+		/* xstrdup() required below so that we don't hack up the environment */
+		manpathlist = xstrdup (manpathlist);
+		if ( manpathlist[0] == ':' ) {
+			if (!quiet) error (0, 0, 
+				_( "warning: $MANPATH set, prepending %s"), CONFIG_FILE);
+			manpathlist = strappend ( guess_manpath (systems)
+						, add_system_manpath (systems, manpathlist)
+						, NULL);
+		} else if ( manpathlist[strlen(manpathlist)-1] == ':' ) {
+			if (!quiet) error (0, 0, 
+				_( "warning: $MANPATH set, appending %s"), CONFIG_FILE);
+			manpathlist = strappend ( add_system_manpath (systems, manpathlist)
+						, guess_manpath (systems)
+						, NULL);
+		} else if ( (pos = strstr(manpathlist,"::")) ) {
+			*(pos++) = '\0';
+			if (!quiet) error (0, 0, 
+				_( "warning: $MANPATH set, inserting '%s'"), CONFIG_FILE);
+			manpathlist = strappend ( add_system_manpath (systems, manpathlist)
+						, ":" , guess_manpath (systems)
+						, add_system_manpath (systems, pos)
+						, NULL);
+		} else {
+			if (!quiet) error (0, 0, 
+				_( "warning: $MANPATH set, ignoring %s"), CONFIG_FILE);
+			manpathlist = add_system_manpath (systems, manpathlist);
+		}
+	} else {
+		manpathlist = guess_manpath (systems);
+	}
+	return manpathlist;
+}
+
+#ifdef SECURE_MAN_UID
+extern uid_t ruid;			/* initial real user id */
+extern uid_t euid;			/* initial effective user id */
+#endif /* SECURE_MAN_UID */
+/* create the catman hierarchy if it doesn't exist */
+void
+mkcatdirs( DIRLIST *dlp)
+{
+	char manname[PATH_MAX+6];
+	char catname[PATH_MAX+6];
+
+	if ( dlp->bin ) {
+		int oldmask=umask(022);
+		/* first the base catdir */
+		if ( is_directory( dlp->bin) != 1) {
+			regain_effective_privs();
+			if ( mkdir( dlp->bin, S_ISGID|0755) < 0 ) {
+				if (!quiet)
+					error (0, 0, _( "warning: cannot create catdir %s"), dlp->bin);
+				if (debug)
+					fprintf (stderr, "warning: cannot create catdir %s\n", dlp->bin);
+			} else if (debug)
+				fprintf (stderr, "created base catdir %s\n", dlp->bin);
+#ifdef SECURE_MAN_UID
+			if ( ruid == 0 )
+				chown( dlp->bin, euid, 0);
+#endif /* SECURE_MAN_UID */
+			drop_effective_privs();
+		}
+		/* then the hierarchy */
+		sprintf( catname, "%s/cat1", dlp->bin);
+		if ( is_directory( dlp->bin) == 1) {
+			int j;
+			regain_effective_privs();
+			if (debug)	fprintf (stderr, 
+				"creating catdir hierarchy %s  ", dlp->bin);
+			for (j=1; j<10; j++) {
+				sprintf( catname, "%s/cat%d", dlp->bin, j);
+				sprintf( manname, "%s/man%d", dlp->mandir, j);
+				if (( is_directory( manname) == 1)
+				&& (  is_directory( catname) != 1)) {
+					if ( mkdir( catname, S_ISGID|0755) < 0 ) {
+						if (!quiet)
+							error (0, 0, _( "warning: cannot create catdir %s"), catname);
+						if (debug)
+							fprintf (stderr, "warning: cannot create catdir %s\n", catname);
+					} else if (debug)
+						fprintf (stderr, " cat%d", j);
+#ifdef SECURE_MAN_UID
+					if ( ruid == 0 )
+						chown( catname, euid, 0);
+#endif /* SECURE_MAN_UID */
+				}
+			}
+			if (debug)
+				fprintf (stderr, "\n");
+			drop_effective_privs();
+		}
+		umask(oldmask);
+	}
+}
+
+extern char *program_name;
 /* Parse the manpath.config file, extracting appropriate information. */
-static __inline__ int get_dirlist (void)
+DIRLIST * add_2_dirlist ( FILE *config, DIRLIST *dlp)
 {
 	char *bp;
 	char buf[BUFSIZ];
-	DIRLIST *dlp = list;
-	FILE *config;
 	int converted;
 
-	if ((config = fopen (CONFIG_FILE, "r")) == NULL)
-		error (FAIL, 0,
-		       CATGETS(manp_OPEN_CONF, "can't open the manpath configuration file %s"),
-		       CONFIG_FILE);
-
 	while ( (bp = fgets (buf, BUFSIZ, config)) ) {
 
 		while ( isspace(*bp) )
 			bp++;
-			
+
+		/* leave last element empty! */
+		if ( dlp >= &list[MAXDIRS-1] ) {
+			error (0, 0, _( "too many entries in config file"));
+			gripe_reading_mp_config (CONFIG_FILE);
+			return dlp;
+		}
+
 		if (*bp == '#' || *bp == '\0') {
 			continue;
+		} else if (strncmp(bp, "NO", 2) == 0) {
+			continue;	/* mach any word starting with NO */
 		} else if (sscanf(bp, "MANBIN %*s") == 1) {
 			continue;
 		} else if (sscanf(bp, "MANDATORY_MANPATH %s", dlp->mandir) == 1) {
@@ -426,38 +554,58 @@ static __inline__ int get_dirlist (void)
 		} else if ( (converted = 
 		  sscanf(bp, "MANDB_MAP %s %s", dlp->mandir, dlp->bin)) > 0) {
 		  	dlp->mandatory = MANDB_MAP;
-		  	if (is_directory(dlp->mandir) != 1)
+		  	if (is_directory(dlp->mandir) != 1) {
 		  		continue;
+			}
 		  	if (converted == 1) {
 		  		(void) strcpy(dlp->bin, dlp->mandir);
+#ifdef USE_FSSTND
 		  	} else if (converted == 2 && 
 		  	         strcmp(dlp->bin, "FSSTND") == 0) {
 		  	        char *fsstnd_catpath;
 		  		(void) strcpy(dlp->bin, 
 		  		              fsstnd_catpath = fsstnd(dlp->mandir));
 		  		free(fsstnd_catpath);
+#endif /* USE_FSSTND */
 		  	}
 		  	if (debug)
 		  		fprintf (stderr, 
 		  		  "found global mandir %s mapped to catdir %s\n",
 		  		  dlp->mandir, dlp->bin);
+			/* create the catman hierarchy if it doesn't exist */
+			if ( strcmp(program_name,"mandb") == 0)
+				mkcatdirs( dlp);
 		  	dlp++;
 	 	} else {
-			error (0, 0,
-			       CATGETS(manp_PARSE_DIRS,
-				       "can't parse directory list `%s'"),
-			       bp);
+			error (0, 0, _( "can't parse directory list `%s'"), bp);
 			gripe_reading_mp_config (CONFIG_FILE);
 		}
 	}
 
-	dlp->bin[0] = dlp->mandir[0] = '\0';
-	dlp->mandatory = 0;
+	return dlp;
+}
+int get_dirlist (void)
+{
+	char *home;
+	FILE *config;
+	DIRLIST *dlp = list;
+
+	if ((config = fopen (CONFIG_FILE, "r")) == NULL)
+		error (FAIL, 0, _( "can't open the manpath configuration file %s"), CONFIG_FILE);
+	dlp = add_2_dirlist (config, dlp);
 	fclose(config);
 
+	if ((home = xstrdup( getenv ("HOME"))) )
+		if ((config = fopen ( strappend ( home, "/.manpath", NULL), "r")) != NULL) {
+			dlp = add_2_dirlist (config, dlp);
+			fclose(config);
+		}
+	dlp->bin[0] = dlp->mandir[0] = '\0';
+	dlp->mandatory = 0;
 	return 0;
 }
 
+
 /*
  * Construct the default manpath.  This picks up mandatory manpaths
  * only.
@@ -490,10 +638,7 @@ static char *def_path (int flag)
 		if (status < 0) {
 			gripe_stat_file(dlp->mandir);
 		} else if (status == 0 && !quiet) {
-			error (0, 0,
-			       CATGETS(manp_MISS_DIR,
-				       "warning: mandatory directory %s doesn't exist"),
-			       dlp->mandir);
+			error (0, 0, _( "warning: mandatory directory %s doesn't exist"), dlp->mandir);
 		} else if (status == 1){
 			(void) strcat(manpathlist, dlp->mandir);
 			(void) strcat(manpathlist, ":"); 
@@ -524,13 +669,15 @@ static __inline__ char *get_manpath (char *path)
 	char *manpathlist;
 	DIRLIST *dlp;
 
-	tmppath = xstrdup (path);
-
-	for (p = tmppath;; p = end + 1){
+	for (end = p = tmppath = xstrdup (path); end; p = end + 1){
 	
 		if ((end = strchr (p, ':')))
 			*end = '\0';
 
+		/* don't do this for current dir ("." or empty entry in PATH) */
+		if ((*p == '\0') || (strcmp( p, ".") == 0))
+			continue;
+
 		if (debug)
 			fprintf (stderr, "\npath directory %s ", p);
 
@@ -569,9 +716,7 @@ static __inline__ char *get_manpath (char *path)
 				fprintf (stderr, "and doesn't have man or MAN subdirectories\n");
 
 	found:
-
-		  if (!end)
-			break;
+		;
 	}
 
 	free(tmppath);
@@ -700,30 +845,31 @@ static __inline__ char **add_dir_to_path_list (char **mp, char *p)
 #else /* no GETCWD */
 			if ( !getwd(cwd) )
 #endif /* HAVE_GETCWD */
-				error (FATAL, errno,
-				       CATGETS(manp_CWD, "can't determine current directory"));
+				error (FATAL, errno, _( "can't determine current directory"));
 			(void) strcat(wd, "/");
 			(void) strcat(wd, p);
 			*mp = xstrdup(wd);
 		} else 
 			*mp = xstrdup (p);
 
+#ifdef MAN_TREE_COMPONENT
 		if (!strstr(*mp, "/man") && !strstr(*mp, "/MAN")) {
 
 			/* not really a man tree after all */
 
 			if (!quiet)
-				error (0, 0,
-				       CATGETS(manp_NO_TREE, "warning: %s does not have a man tree component"), 
-				       *mp);
+				error (0, 0, _( "warning: %s does not have a man tree component"), *mp);
 			free(*mp);
 		} else {
+#endif /* MAN_TREE_COMPONENT */
 			if (debug)
 				fprintf (stderr, 
 				  "adding %s to manpathlist\n", 
 				  *mp);
 			mp++;
+#ifdef MAN_TREE_COMPONENT
 		}
+#endif /* MAN_TREE_COMPONENT */
 	}
 	return mp;
 }
@@ -810,6 +956,10 @@ char *global_catpath(char *name)
 				(void) strcat (catpath, "cat");
 			}
 			(void) strcat (catpath, suffix);
+#if 0
+			if (debug)
+			fprintf(stderr, "global_catpath: %s\n", catpath);
+#endif
 		  	return catpath;
 		}
 	}
@@ -838,10 +988,7 @@ static __inline__ char *fsstnd(char *path)
 	
 	if (strncmp(path, MAN_ROOT, sizeof MAN_ROOT - 1) != 0) {
 		if (!quiet)
-			error (0, 0,
-			       CATGETS(manp_PREFIX,
-				       "warning: %s does not begin with %s"),
-			       path, MAN_ROOT);
+			error (0, 0, _( "warning: %s does not begin with %s"), path, MAN_ROOT);
 		return xstrdup(path);
 	}
 	/* get rid of initial "/usr" */
