@@ -55,8 +55,12 @@ extern char *strrchr();
 #  endif /* HAVE_NDIR_H */
 #endif /* HAVE_DIRENT_H  */
 
-#define NLS_SET	straySet
-#include "nls/nls.h"
+#ifdef HAVE_CANONICALIZE_FILE_NAME
+extern char *canonicalize_file_name __P ((__const char *__name));
+#endif
+
+#include <libintl.h>
+#define _(String) gettext (String)
 
 #include "manconfig.h"
 #include "libdb/mydbm.h"
@@ -76,11 +80,15 @@ static __inline__ int check_for_stray(void)
 	struct dirent *catlist;
 	size_t lenman, lencat;
 	int strays = 0;
+#ifdef HAVE_CANONICALIZE_FILE_NAME
+	/* no PATH_MAX then */
+	char *fullpath;
+#else
+	char fullpath[PATH_MAX];
+#endif
 	
 	if ( !(cdir = opendir(catdir))) {
-		error (0, errno,
-		       CATGETS(stray_OPENDIR, "can't search directory %s"),
-		       catdir);
+		error (0, errno, _( "can't search directory %s"), catdir);
 		return 0;
 	}
 
@@ -108,10 +116,8 @@ static __inline__ int check_for_stray(void)
 
 		ext = strrchr(mandir, '.');
 		if (!ext) {
-			if (!quiet)
-				error (0, 0,
-				       CATGETS(stray_BOGUS, "warning: %s: ignoring bogus filename"),
-				       catdir);
+			if (quiet < 2)
+				error (0, 0, _( "warning: %s: ignoring bogus filename"), catdir);
 			continue;
 			
 #if defined(COMP_SRC) || defined(COMP_CAT)
@@ -136,10 +142,8 @@ static __inline__ int check_for_stray(void)
 		/* check for bogosity */
 		
 		if (!ext || strncmp(ext + 1, section, strlen(section)) != 0) {
-		  	if (!quiet)
-				error (0, 0,
-				       CATGETS(stray_BOGUS, "warning: %s: ignoring bogus filename"),
-			  	       catdir);
+			if (quiet < 2)
+				error (0, 0, _( "warning: %s: ignoring bogus filename"), catdir);
 			free(section);
 			continue;
 		}
@@ -164,7 +168,7 @@ static __inline__ int check_for_stray(void)
 			found = 0;
 			
 		if (!found) { 
-			char *filter, *whatis;
+			char *filter=0, *whatis=0;
 			struct mandata *exists;
 
 			/* we have a straycat. Need to filter it and get
@@ -214,28 +218,45 @@ static __inline__ int check_for_stray(void)
 				filter = strappend(NULL, COL " -bx < ",
 						   catdir, " > ", temp_name,
 						   NULL);
-					
-			if (do_system(filter) != 0) {
-				remove(temp_name);
-				if (debug)
+                                                                               
+#ifdef HAVE_CANONICALIZE_FILE_NAME                                                                  
+			if ( !(fullpath = canonicalize_file_name(catdir)) ) {
+				if (quiet < 2)
+					error (0, errno, _( "can't resolve %s"), catdir);
+			} else
+#else                  
+			if (realpath(catdir, fullpath) == NULL) {
+				if (quiet < 2)
+					if (errno == ENOENT)
+						error (0, 0, _( "warning: %s is a dangling symlink"), fullpath);
+					else
+						error (0, errno, _( "can't resolve %s"), fullpath);
+				}
+			} else 
+#endif
+			{
+#ifdef HAVE_CANONICALIZE_FILE_NAME                                                                  
+				free(fullpath);                                                      
+#endif
+				if (do_system(filter) != 0) {
+					remove(temp_name);
 					perror(filter);
-				exit (CHILD_FAIL);
+					exit (CHILD_FAIL);
+				}
+
+				strays++;
+				
+				if ( (whatis = find_name(temp_name, basename(catdir),
+							 CATPAGE)) == NULL)
+					if (quiet < 2)
+						error (0, 0, _( "warning: %s: whatis parse for %s(%s) failed"),
+						       catdir, basename(mandir), info.sec);
+
+				(void) splitline(whatis, &info, basename(mandir));
 			}
 			
 			free(filter);
 
-			strays++;
-			
-			if ( (whatis = find_name(temp_name, basename(catdir),
-						 CATPAGE)) == NULL)
-				if (!quiet)
-					error (0, 0,
-					       CATGETS(stray_NO_WHATIS,
-						       "warning: %s: whatis parse for %s(%s) failed"),
-					       catdir, basename(mandir), info.sec);
-
-			(void) splitline(whatis, &info, basename(mandir));
-
 			if (whatis)
 				free(whatis);
 		} 
@@ -253,15 +274,12 @@ static int open_catdir(void)
 	int strays = 0;
 
 	if ( !(cdir = opendir(catdir))) {
-		error (0, errno,
-		       CATGETS(stray_OPENDIR, "can't search directory %s"),
-		       catdir);
+		error (0, errno, _( "can't search directory %s"), catdir);
 		return 0;
 	}
 
 	if (!quiet)
-		printf(CATGETS(stray_CHECK,
-			       "Checking for stray cats under %s...\n"), catdir);
+		printf(_( "Checking for stray cats under %s...\n"), catdir);
 
 	catdir = strappend(catdir, "/", NULL);
 	mandir = strappend(mandir, "/", NULL);
@@ -294,13 +312,20 @@ int straycats(char *manpath)
 	char *catpath;
 	int strays;
 
-	if (!temp_name)
-		temp_name = tempnam(NULL, "zcat");
+	if (!temp_name) {
+		int fd;
+
+		if ( !(temp_name=xstrdup("/tmp/zcatXXXXXX")) )
+			return 0;
+		if ( (fd = mkstemp(temp_name)) == -1) {
+			error (0, errno, _( "warning: can't create temp file %s"), temp_name);
+			return 0;
+		}
+		close( fd);
+	}
 	
 	if ( !(dbf = MYDBM_RWOPEN(database)) || dbver_rd(dbf)) {
-		error (0, errno,
-		       CATGETS(stray_UPDATE_DB, "warning: can't update index cache %s"),
-		       database);
+		error (0, errno, _( "warning: can't update index cache %s"), database);
 		return 0;
 	}
 	
@@ -332,5 +357,6 @@ int straycats(char *manpath)
 
 	MYDBM_CLOSE(dbf);
 	remove (temp_name);
+	free(temp_name); temp_name = NULL;
 	return strays;
 }
